# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from typing import List

import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import (
    KeymapTranslationHandler,
    TranslationHandlerGroup,
)


__all__ = ["ObservationFromCurrentLocation"]


class ObservationFromCurrentLocation(TranslationHandlerGroup):
    """
    Includes the current biome, how likely rain and snow are there, as well as the current light level, how bright the
    sky is, and if the player can see the sky.

    Also includes x, y, z, roll, and pitch
    """

    def xml_template(self) -> str:
        return str("""<ObservationFromFullStats/>""")

    def to_string(self) -> str:
        return "location_stats"

    def __init__(self):
        super(ObservationFromCurrentLocation, self).__init__(
            handlers=[
                _SunBrightnessObservation(),
                _SkyLightLevelObservation(),
                _LightLevelObservation(),
                _CanSeeSkyObservation(),
                _BiomeRainfallObservation(),
                _BiomeTemperatureObservation(),
                _IsRainingObservation(),
                # TODO _BiomeNameObservation(),
                _BiomeIDObservation(),
                _PitchObservation(),
                _YawObservation(),
                _SeaLevelObservation(),
                _PosObservation(),
            ]
        )


class _PosObservation(KeymapTranslationHandler):
    def __init__(self):
        keys = ["xpos", "ypos", "zpos"]
        space = spaces.Box(low=-640000.0, high=640000.0, shape=(3,), dtype=np.float32)
        default = np.array([0.0, 0.0, 0.0])
        super().__init__(
            hero_keys=keys,
            univ_keys=keys,
            space=space,
            default_if_missing=default,
            to_string="pos",
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.array(
            [hero_dict["xpos"], hero_dict["ypos"], hero_dict["zpos"]],
            dtype=self.space.dtype,
        )


class _FullStatsObservation(KeymapTranslationHandler):
    def to_hero(self, x) -> int:
        for key in self.hero_keys:
            x = x[key]
        return x

    def __init__(self, key_list: List[str], space=None, default_if_missing=None):
        if space is None:
            if "achievement" == key_list[0]:
                space = spaces.Box(low=0, high=1, shape=(), dtype=np.int)
            else:
                space = spaces.Box(low=0, high=np.inf, shape=(), dtype=np.int)
        if default_if_missing is None:
            default_if_missing = np.zeros((), dtype=np.float)

        super().__init__(
            hero_keys=key_list,
            univ_keys=key_list,
            space=space,
            default_if_missing=default_if_missing,
        )

    def xml_template(self) -> str:
        return str("""<ObservationFromFullStats/>""")


class _SunBrightnessObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["sun_brightness"],
            space=spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32),
            default_if_missing=0.94,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _SkyLightLevelObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["sky_light_level"],
            space=spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32),
            default_if_missing=0.71,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _PitchObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["pitch"],
            space=spaces.Box(low=-180.0, high=180.0, shape=(1,), dtype=np.float32),
            default_if_missing=0.0,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _YawObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["yaw"],
            space=spaces.Box(low=-180.0, high=180.0, shape=(1,), dtype=np.float32),
            default_if_missing=0.0,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _BiomeIDObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["biome_id"],
            space=spaces.Box(low=0, high=167, shape=(), dtype=int),
            default_if_missing=0,
        )


class _BiomeRainfallObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["biome_rainfall"],
            space=spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32),
            default_if_missing=0.5,
        )

    def to_string(self) -> str:
        return "rainfall"

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _BiomeTemperatureObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["biome_temperature"],
            space=spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32),
            default_if_missing=0.5,
        )

    def to_string(self) -> str:
        return "temperature"

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _SeaLevelObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["sea_level"],
            space=spaces.Box(low=0.0, high=255, shape=(1,), dtype=np.float32),
            default_if_missing=63,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _LightLevelObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["light_level"],
            space=spaces.Box(low=0.0, high=15, shape=(1,), dtype=np.float32),
            default_if_missing=15,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _IsRainingObservation(_FullStatsObservation):
    def __init__(self):
        super().__init__(
            key_list=["is_raining"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=int),
            default_if_missing=0,
        )


class _CanSeeSkyObservation(_FullStatsObservation):
    def to_hero(self, x):
        for key in self.hero_keys:
            x = x[key]
        return np.eye(2)[x]

    def __init__(self):
        super().__init__(
            key_list=["can_see_sky"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=int),
            default_if_missing=1,
        )



import os
from typing import List, Optional

import jinja2
from lxml import etree

from . import spaces as spaces
from .handler import Handler
from .handlers.translation import TranslationHandler
from .bridge.mc_instance.instance import MALMO_VERSION


MISSION_TEMPLATE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "mc_meta", "minedojo_mission.xml.j2"
)


class SimSpec:
    MALMO_VERSION = MALMO_VERSION

    def __init__(
        self,
        *,
        sim_name: str,
        agent_count: int = 1,
        obs_handlers: List[TranslationHandler],
        action_handlers: List[TranslationHandler],
        agent_handlers: List[Handler],
        agent_start_handlers: List[Handler],
        server_initial_conditions_handlers: List[Handler],
        world_generator_handlers: List[Handler],
        server_decorator_handlers: List[Handler],
        server_quit_handlers: List[Handler],
        seed: Optional[int] = None,
    ):
        self._sim_name = sim_name
        assert agent_count == 1, "TODO"
        self._agent_count = agent_count
        self._agent_names = [f"agent_{role}" for role in range(agent_count)]

        self._obs_handlers = obs_handlers
        self._action_handlers = action_handlers
        self._agent_handlers = agent_handlers
        self._server_initial_conditions_handlers = server_initial_conditions_handlers
        self._world_generator_handlers = world_generator_handlers
        self._server_decorator_handlers = server_decorator_handlers
        self._server_quit_handlers = server_quit_handlers
        self._agent_start_handlers_list = [
            agent_start_handlers for _ in range(agent_count)
        ]

        # check that the obs/action have no duplicate to_strings
        assert len([o.to_string() for o in self._obs_handlers]) == len(
            set([o.to_string() for o in self._obs_handlers])
        )
        assert len([a.to_string() for a in self._action_handlers]) == len(
            set([a.to_string() for a in self._action_handlers])
        )
        self._observation_space = self.create_observation_space()
        self._action_space = self.create_action_space()
        self._observation_space.seed(seed)
        self._action_space.seed(seed)

        self._episode_id = None

    @property
    def observation_space(self) -> spaces.Dict:
        return self._observation_space

    @property
    def action_space(self) -> spaces.Dict:
        return self._action_space

    @property
    def observables(self):
        return self._obs_handlers

    @property
    def actionables(self):
        return self._action_handlers

    def _singlify(self, space: spaces.Dict):
        if self._agent_count == 1:
            return space.spaces[self._agent_names[0]]
        else:
            return space

    def create_observation_space(self):
        return self._singlify(
            spaces.Dict(
                {
                    agent: spaces.Dict(
                        {o.to_string(): o.space for o in self._obs_handlers}
                    )
                    for agent in self._agent_names
                }
            )
        )

    def create_action_space(self):
        return self._singlify(
            spaces.Dict(
                {
                    agent: spaces.Dict(
                        {a.to_string(): a.space for a in self._action_handlers}
                    )
                    for agent in self._agent_names
                }
            )
        )

    def to_xml(self, episode_id: str) -> str:
        """
        Gets the XML by templating mission.xml.j2 using Jinja
        """
        self._episode_id = episode_id
        with open(MISSION_TEMPLATE, "rt") as fh:
            var_dict = {}
            for attr_name in dir(self):
                if "to_xml" not in attr_name:
                    var_dict[attr_name] = getattr(self, attr_name)

            env = jinja2.Environment(undefined=jinja2.StrictUndefined)
            template = env.from_string(fh.read())

        xml = template.render(var_dict)
        # Now do one more pretty printing

        xml = etree.tostring(
            etree.fromstring(xml.encode("utf-8")), pretty_print=True
        ).decode("utf-8")
        return xml

    @staticmethod
    def get_consolidated_xml(handlers: List[Handler]) -> List[str]:
        """Consolidates duplicate XML representations from the handlers.

        Deduplication happens by first getting all handler.xml() strings
        of the handlers, and then converting them into etrees. After that we check
        if there are any top level elements that are duplicated and pick the first of them
        to retain. We then convert the remaining etrees back into strings and join them with new lines.

        Args:
            handlers (List[Handler]): A list of handlers to consolidate.

        Returns:
            str: The XML
        """
        handler_xml_strs = [handler.xml() for handler in handlers]

        if not handler_xml_strs:
            return ""

        # TODO: RAISE VALID XML ERROR. FOR EASE OF USE
        trees = [etree.fromstring(xml) for xml in handler_xml_strs if xml != ""]
        consolidated_trees = {tree.tag: tree for tree in trees}.values()

        return [
            etree.tostring(t, pretty_print=True).decode("utf-8")
            for t in consolidated_trees
        ]



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from typing import Optional

from minedojo.sim.handlers.agent.action import ItemListAction


class PlaceBlock(ItemListAction):
    """
    An action handler for placing a specific block
    """

    def to_string(self):
        return "place"

    def xml_template(self) -> str:
        return str("<PlaceCommands/>")

    def __init__(self, blocks: list, _other=Optional[str], _default=Optional[str]):
        """
        Initializes the space of the handler to be one for each item in the list
        Requires 0th item to be 'none' and last item to be 'other' corresponding to
        no-op and non-listed item respectively
        """
        self._items = blocks
        self._command = "place"
        kwargs = {}
        if _other is not None:
            kwargs["_other"] = _other
        if _default is not None:
            kwargs["_default"] = _default
        super().__init__(self._command, self._items, **kwargs)
        self._prev_inv = None

    def from_universal(self, obs):
        try:
            for action in obs["custom_action"]["actions"].keys():
                try:
                    if int(action) == -99 and self._prev_inv is not None:

                        item_name = self._prev_inv[int(-10 + obs["hotbar"])][
                            "name"
                        ].split("minecraft:")[-1]
                        if item_name not in self._items:
                            raise ValueError()
                        else:
                            return item_name
                except ValueError:
                    return self._other
        except TypeError:
            print("Saw a type error in PlaceBlock")
            raise TypeError
        except KeyError:
            return self._default
        finally:
            try:
                self._prev_inv = obs["slots"]["gui"]["slots"]
            except KeyError:
                self._prev_inv = None

        return self._default



from typing import Union
from copy import deepcopy
from collections import deque

import gym
import numpy as np

from ..utils import *
from ...sim import MineDojoSim
from ....sim import spaces as spaces
from ....sim.mc_meta import mc as MC


ALL_ITEMS = np.array(MC.ALL_ITEMS)


class DeltaInventoryWrapper(gym.Wrapper):
    def __init__(
        self,
        env: Union[MineDojoSim, gym.Wrapper],
        op_action_idx: int = 5,
        craft_action_idx: int = 4,
        craft_arg_idx: int = 6,
        n_increased: int = 1,
        n_decreased: int = 4,
        default_item_name: str = "air",
    ):
        assert "inventory" in env.observation_space.keys()
        assert "masks" in env.observation_space.keys()
        assert "craft_smelt" in env.observation_space["masks"].keys()
        assert isinstance(
            env.action_space, spaces.MultiDiscrete
        ), "please use this wrapper with `NNActionSpaceWrapper!`"
        assert (
            len(env.action_space.nvec) == 8
        ), "please use this wrapper with `NNActionSpaceWrapper!`"
        assert op_action_idx < len(env.action_space.nvec)
        assert craft_arg_idx < len(env.action_space.nvec)
        super().__init__(env=env)
        obs_space = env.observation_space
        obs_space["delta_inv"] = spaces.Dict(
            {
                "inc_name_by_craft": spaces.Text(shape=(n_increased,)),
                "inc_quantity_by_craft": spaces.Box(
                    low=0, high=64, shape=(n_increased,), dtype=np.float32
                ),
                "dec_name_by_craft": spaces.Text(shape=(n_decreased,)),
                "dec_quantity_by_craft": spaces.Box(
                    low=0, high=64, shape=(n_decreased,), dtype=np.float32
                ),
                "inc_name_by_other": spaces.Text(shape=(n_increased,)),
                "inc_quantity_by_other": spaces.Box(
                    low=0, high=64, shape=(n_increased,), dtype=np.float32
                ),
                "dec_name_by_other": spaces.Text(shape=(n_decreased,)),
                "dec_quantity_by_other": spaces.Box(
                    low=0, high=64, shape=(n_decreased,), dtype=np.float32
                ),
            }
        )
        self.observation_space = obs_space
        self._default_item_name = default_item_name
        self._n_increased = n_increased
        self._n_decreased = n_decreased
        self._op_action_idx = op_action_idx
        self._craft_action_idx = craft_action_idx
        self._craft_arg_idx = craft_arg_idx

        self._recipes = get_recipes_matrix()
        self._prev_inventory = None
        self._prev_mask = None
        # note that there is two-step lag between executing a crafting action
        # and that item really goes into the inventory
        # so we use a deque with len=3 to cache the history action
        self._prev_craft_actions = deque(maxlen=1)

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)
        self._prev_inventory = deepcopy(observation["inventory"])
        self._prev_mask = deepcopy(observation["masks"]["craft_smelt"])
        self._prev_craft_actions = deque(maxlen=1)
        return self.observation(observation, None)

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        new_obs = self.observation(observation, action)
        self._prev_inventory = deepcopy(observation["inventory"])
        self._prev_mask = deepcopy(observation["masks"]["craft_smelt"])
        return new_obs, reward, done, info

    def observation(self, observation, action):
        if action is None:
            # first step
            delta_obs = {
                "inc_name_by_craft": np.array(
                    [self._default_item_name] * self._n_increased
                ),
                "inc_quantity_by_craft": np.zeros(
                    (self._n_increased,), dtype=np.float32
                ),
                "dec_name_by_craft": np.array(
                    [self._default_item_name] * self._n_decreased
                ),
                "dec_quantity_by_craft": np.zeros(
                    (self._n_decreased,), dtype=np.float32
                ),
                "inc_name_by_other": np.array(
                    [self._default_item_name] * self._n_increased
                ),
                "inc_quantity_by_other": np.zeros(
                    (self._n_increased,), dtype=np.float32
                ),
                "dec_name_by_other": np.array(
                    [self._default_item_name] * self._n_decreased
                ),
                "dec_quantity_by_other": np.zeros(
                    (self._n_decreased,), dtype=np.float32
                ),
            }
        else:
            new_craft_idx = (
                None
                if action[self._op_action_idx] != self._craft_action_idx
                else action[self._craft_arg_idx]
            )
            self._prev_craft_actions.append(new_craft_idx)
            craft_idx = self._prev_craft_actions[0]
            cur_inv_vector = get_inventory_vector(observation["inventory"])[0]
            pre_inv_vector = get_inventory_vector(self._prev_inventory)[0]
            delta_inv_vector = cur_inv_vector - pre_inv_vector
            increment_indices = np.flatnonzero(delta_inv_vector > 0)[
                : self._n_increased
            ]
            decrement_indices = np.flatnonzero(delta_inv_vector < 0)[
                : self._n_decreased
            ]

            delta_obs = {}
            if len(increment_indices) == 0:
                # no increment
                delta_obs.update(
                    {
                        "inc_name_by_craft": np.array(
                            [self._default_item_name] * self._n_increased
                        ),
                        "inc_quantity_by_craft": np.zeros(
                            (self._n_increased,), dtype=np.float32
                        ),
                        "inc_name_by_other": np.array(
                            [self._default_item_name] * self._n_increased
                        ),
                        "inc_quantity_by_other": np.zeros(
                            (self._n_increased,), dtype=np.float32
                        ),
                    }
                )
            else:
                # "craft" is not selected!
                if craft_idx is None:
                    delta_obs.update(
                        {
                            "inc_name_by_craft": np.array(
                                [self._default_item_name] * self._n_increased
                            ),
                            "inc_quantity_by_craft": np.zeros(
                                (self._n_increased,), dtype=np.float32
                            ),
                            "inc_name_by_other": np.concatenate(
                                (
                                    ALL_ITEMS[increment_indices],
                                    np.array(
                                        [self._default_item_name]
                                        * (self._n_increased - len(increment_indices))
                                    ),
                                )
                            ),
                            "inc_quantity_by_other": np.concatenate(
                                (
                                    delta_inv_vector[increment_indices],
                                    np.zeros(
                                        shape=(
                                            self._n_increased - len(increment_indices),
                                        )
                                    ),
                                )
                            ).astype(np.float32),
                        }
                    )
                else:
                    item_name_to_craft = MC.ALL_CRAFT_SMELT_ITEMS[craft_idx]
                    item_idx_to_craft = MC.ALL_ITEMS.index(item_name_to_craft)
                    if (
                        bool(self._prev_mask[craft_idx]) is True
                        and item_idx_to_craft in increment_indices
                    ):
                        delta_obs.update(
                            {
                                "inc_name_by_craft": np.concatenate(
                                    (
                                        ALL_ITEMS[increment_indices],
                                        np.array(
                                            [self._default_item_name]
                                            * (
                                                self._n_increased
                                                - len(increment_indices)
                                            )
                                        ),
                                    )
                                ),
                                "inc_quantity_by_craft": np.concatenate(
                                    (
                                        delta_inv_vector[increment_indices],
                                        np.zeros(
                                            shape=(
                                                self._n_increased
                                                - len(increment_indices),
                                            )
                                        ),
                                    )
                                ).astype(np.float32),
                                "inc_name_by_other": np.array(
                                    [self._default_item_name] * self._n_increased
                                ),
                                "inc_quantity_by_other": np.zeros(
                                    (self._n_increased,), dtype=np.float32
                                ),
                            }
                        )
                    else:
                        delta_obs.update(
                            {
                                "inc_name_by_craft": np.array(
                                    [self._default_item_name] * self._n_increased
                                ),
                                "inc_quantity_by_craft": np.zeros(
                                    (self._n_increased,), dtype=np.float32
                                ),
                                "inc_name_by_other": np.concatenate(
                                    (
                                        ALL_ITEMS[increment_indices],
                                        np.array(
                                            [self._default_item_name]
                                            * (
                                                self._n_increased
                                                - len(increment_indices)
                                            )
                                        ),
                                    )
                                ),
                                "inc_quantity_by_other": np.concatenate(
                                    (
                                        delta_inv_vector[increment_indices],
                                        np.zeros(
                                            shape=(
                                                self._n_increased
                                                - len(increment_indices),
                                            )
                                        ),
                                    )
                                ).astype(np.float32),
                            }
                        )
            if len(decrement_indices) == 0:
                # no decrement
                delta_obs.update(
                    {
                        "dec_name_by_craft": np.array(
                            [self._default_item_name] * self._n_decreased
                        ),
                        "dec_quantity_by_craft": np.zeros(
                            (self._n_decreased,), dtype=np.float32
                        ),
                        "dec_name_by_other": np.array(
                            [self._default_item_name] * self._n_decreased
                        ),
                        "dec_quantity_by_other": np.zeros(
                            (self._n_decreased,), dtype=np.float32
                        ),
                    }
                )
            else:
                # "craft" is not selected!
                if craft_idx is None:
                    delta_obs.update(
                        {
                            "dec_name_by_craft": np.array(
                                [self._default_item_name] * self._n_decreased
                            ),
                            "dec_quantity_by_craft": np.zeros(
                                (self._n_decreased,), dtype=np.float32
                            ),
                            "dec_name_by_other": np.concatenate(
                                (
                                    ALL_ITEMS[decrement_indices],
                                    np.array(
                                        [self._default_item_name]
                                        * (self._n_decreased - len(decrement_indices))
                                    ),
                                )
                            ),
                            "dec_quantity_by_other": abs(
                                np.concatenate(
                                    (
                                        delta_inv_vector[decrement_indices],
                                        np.zeros(
                                            (
                                                self._n_decreased
                                                - len(decrement_indices),
                                            )
                                        ),
                                    )
                                ).astype(np.float32)
                            ),
                        }
                    )
                else:
                    ingredients_indices = np.flatnonzero(self._recipes[craft_idx] > 0)
                    if bool(self._prev_mask[craft_idx]) is True and set(
                        ingredients_indices
                    ).issubset(set(decrement_indices)):
                        delta_obs.update(
                            {
                                "dec_name_by_craft": np.concatenate(
                                    (
                                        ALL_ITEMS[decrement_indices],
                                        np.array(
                                            [self._default_item_name]
                                            * (
                                                self._n_decreased
                                                - len(decrement_indices)
                                            )
                                        ),
                                    )
                                ),
                                "dec_quantity_by_craft": abs(
                                    np.concatenate(
                                        (
                                            delta_inv_vector[decrement_indices],
                                            np.zeros(
                                                (
                                                    self._n_decreased
                                                    - len(decrement_indices),
                                                )
                                            ),
                                        )
                                    ).astype(np.float32)
                                ),
                                "dec_name_by_other": np.array(
                                    [self._default_item_name] * self._n_decreased
                                ),
                                "dec_quantity_by_other": np.zeros(
                                    (self._n_decreased,), dtype=np.float32
                                ),
                            }
                        )
                    else:
                        delta_obs.update(
                            {
                                "dec_name_by_craft": np.array(
                                    [self._default_item_name] * self._n_decreased
                                ),
                                "dec_quantity_by_craft": np.zeros(
                                    (self._n_decreased,), dtype=np.float32
                                ),
                                "dec_name_by_other": np.concatenate(
                                    (
                                        ALL_ITEMS[decrement_indices],
                                        np.array(
                                            [self._default_item_name]
                                            * (
                                                self._n_decreased
                                                - len(decrement_indices)
                                            )
                                        ),
                                    )
                                ),
                                "dec_quantity_by_other": abs(
                                    np.concatenate(
                                        (
                                            delta_inv_vector[decrement_indices],
                                            np.zeros(
                                                (
                                                    self._n_decreased
                                                    - len(decrement_indices),
                                                )
                                            ),
                                        )
                                    ).astype(np.float32)
                                ),
                            }
                        )
        observation["delta_inv"] = delta_obs
        return observation



from typing import Optional, Union, List, Dict, Tuple

from .base import MetaTaskBase
from ...sim.inventory import InventoryItem
from .utils import simple_inventory_based_check, possess_item_reward


# A human beginner takes 8 hours to defeat the dragon. Each second includes four actions.
TIME_LIMIT = int(8 * 3600 * 4)


class Playthrough(MetaTaskBase):
    """
    Class for the playthrough tasks.
    Args:
        always_night: If ``True``, the world will always be at night.
                Default: ``False``.

        break_speed_multiplier: Controls the speed of breaking blocks. A value larger than 1.0 accelerates the breaking.
                Default: ``1.0``.

        event_level_control: If ``True``, the agent is able to perform high-level controls including place and equip.
                If ``False``, then is keyboard-mouse level control.
                Default: ``True``.

        fast_reset: If ``True``, reset using MC native command `/kill`, instead of waiting for re-generating new worlds.
            Default: ``True``.

            .. warning::
                Side effects:

                1. Changes to the world will not be reset. E.g., if the agent chops lots of trees then calling
                fast reset will not restore those trees.

                2. If you specify agent starting health and food, these specs will only be respected at the first reset
                (i.e., generating a new world) but will not be respected in the following resets (i.e., reset using MC cmds).
                So be careful to use this wrapper if your usages require specific initial health/food.

                3. Statistics/achievements will not be reset. This wrapper will maintain a property ``info_prev_reset``.
                If your tasks use stat/achievements to evaluation, please retrieve this property and compute differences.

        image_size: The size of image observations.

        initial_weather: If not ``None``, specifies the initial weather.
                Can be one of ``"clear"``, ``"normal"``, ``"rain"``, ``"thunder"``.
                Default: ``None``.

        lidar_rays: Defines the directions and maximum distances of the lidar rays if ``use_lidar`` is ``True``.
                If supplied, should be a list of tuple(pitch, yaw, distance).
                Pitch and yaw are in radians and relative to agent looking vector.
                Default: ``None``.

        obtain_dragon_egg_reward: The reward value of obtaining the dragon egg.
                The dragon egg can be solely obtained by successfully defeating the ender dragon.
                So it serves as a proxy for playing through the vanilla game.
                Default: ``10``.

        seed: The seed for an instance's internal generator.
                Default: ``None``.

        sim_name: Name of a simulation instance.
                Default: ``"Playthrough"``.

        start_at_night: If ``True``, the task starts at night.
                Default: ``True``.

        start_food: If not ``None``, specifies initial food condition of the agent.
                Default: ``None``.

        start_health: If not ``None``, specifies initial health condition of the agent.
                Default: ``None``.

        start_position: If not ``None``, specifies the agent's initial location and orientation in the minecraft world.
                Default: ``None``.

        use_lidar: If ``True``, includes lidar in observations.
                Default: ``False``.

        use_voxel: If ``True``, includes voxel in observations.
                Default: ``False``.

        voxel_size: Defines the voxel's range in each axis if ``use_voxel`` is ``True``.
                If supplied, should be a dict with keys ``xmin``, ``xmax``, ``ymin``, ``ymax``, ``zmin``, ``zmax``.
                Each value specifies the voxel size relative to the agent.
                Default: ``None``.

        world_seed: Seed for deterministic world generation
                if ``generate_world_type`` is ``"default"`` or ``"specified_biome"``.
                See `here <https://minecraft.fandom.com/wiki/Seed_(level_generation)>`_ for more details.
                Default: ``None``.
    """

    _prompt_template = "Defeat the Ender Dragon and obtain the trophy dragon egg."

    def __init__(
        self,
        *,
        # ------ reward weights ------
        obtain_dragon_egg_reward: Union[float, int] = 10,
        # ------ initial conditions ------
        start_position: Optional[Dict[str, Union[float, int]]] = None,
        start_health: Optional[float] = None,
        start_food: Optional[int] = None,
        start_at_night: bool = False,
        always_night: bool = False,
        initial_weather: Optional[str] = None,
        initial_inventory: Optional[List[InventoryItem]] = None,
        # ------ sim seed & world seed ------
        seed: Optional[int] = None,
        world_seed: Optional[str] = None,
        # ------ reset mode ------
        fast_reset: bool = True,
        # ------ obs ------
        image_size: Union[int, Tuple[int, int]],
        use_voxel: bool = False,
        voxel_size: Optional[Dict[str, int]] = None,
        use_lidar: bool = False,
        lidar_rays: Optional[List[Tuple[float, float, float]]] = None,
        # ------ event-level action or keyboard-mouse level action ------
        event_level_control: bool = True,
        # ------ misc ------
        break_speed_multiplier: float = 1.0,
        sim_name: str = "Playthrough",
    ):
        success_criteria = [simple_inventory_based_check(name="dragon_egg", quantity=1)]
        reward_fns = [
            possess_item_reward(
                name="dragon_egg", quantity=1, weight=obtain_dragon_egg_reward
            )
        ]

        start_time, allow_time_passage = None, True
        if start_at_night:
            start_time = 13000
        if always_night:
            start_time = 18000
            allow_time_passage = False

        super().__init__(
            fast_reset=fast_reset,
            success_criteria=success_criteria,
            reward_fns=reward_fns,
            seed=seed,
            sim_name=sim_name,
            image_size=image_size,
            use_voxel=use_voxel,
            voxel_size=voxel_size,
            use_lidar=use_lidar,
            lidar_rays=lidar_rays,
            event_level_control=event_level_control,
            initial_inventory=initial_inventory,
            break_speed_multiplier=break_speed_multiplier,
            world_seed=world_seed,
            start_position=start_position,
            initial_weather=initial_weather,
            start_time=start_time,
            allow_time_passage=allow_time_passage,
            start_health=start_health,
            start_food=start_food,
        )

        self.time_limit = TIME_LIMIT
        self._elapsed_steps = None

    @property
    def task_prompt(self) -> str:
        return self._prompt_template

    def reset(self, **kwargs):
        self._elapsed_steps = 0
        return self.env.reset(**kwargs)

    def step(self, action):
        assert (
            self._elapsed_steps is not None
        ), "Cannot call env.step() before calling reset()"
        observation, reward, done, info = self.env.step(action)
        self._elapsed_steps += 1
        if self._elapsed_steps >= self._max_episode_steps:
            info["TimeLimit.truncated"] = not done
            done = True
        return observation, reward, done, info



import os
import pathlib
import pkg_resources
from setuptools import setup, find_packages


PKG_NAME = "minedojo"
VERSION = "0.1"
EXTRAS = {}


def _read_file(fname):
    # this_dir = os.path.abspath(os.path.dirname(__file__))
    # with open(os.path.join(this_dir, fname)) as f:
    with pathlib.Path(fname).open() as fp:
        return fp.read()


def _read_install_requires():
    with pathlib.Path("requirements.txt").open() as fp:
        return [
            str(requirement) for requirement in pkg_resources.parse_requirements(fp)
        ]


def _fill_extras(extras):
    if extras:
        extras["all"] = list(set([item for group in extras.values() for item in group]))
    return extras


setup(
    name=PKG_NAME,
    version=VERSION,
    author=f"MineDojo Team",
    url="http://github.com/MineDojo/MineDojo",
    description="research project",
    long_description=_read_file("README.md"),
    long_description_content_type="text/markdown",
    keywords=["Deep Learning", "Machine Learning"],
    license="MIT License",
    packages=find_packages(include=f"{PKG_NAME}.*"),
    include_package_data=True,
    zip_safe=False,
    install_requires=_read_install_requires(),
    extras_require=_fill_extras(EXTRAS),
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Environment :: Console",
        "Programming Language :: Python :: 3.9",
    ],
)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from typing import Tuple

import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import KeymapTranslationHandler


class POVObservation(KeymapTranslationHandler):
    """
    Handles POV observations.
    """

    def to_string(self):
        return "rgb"

    def __repr__(self):
        result = f"POVObservation(video_resolution={self.video_resolution}"
        if self.include_depth:
            result += ", include_depth=True"
        result += ")"
        result = f"{result}:{self.to_string()}"
        return result

    def xml_template(self) -> str:
        return str(
            """
            <VideoProducer 
                want_depth="{{ include_depth | string | lower }}">
                <Width>{{ video_width }} </Width>
                <Height>{{ video_height }}</Height>
            </VideoProducer>"""
        )

    def __init__(self, video_resolution: Tuple[int, int], include_depth: bool = False):
        self.include_depth = include_depth
        self.video_resolution = video_resolution
        if include_depth:
            space = spaces.Box(0, 255, [4] + list(video_resolution), dtype=np.uint8)
            self.video_depth = 4

        else:
            space = spaces.Box(0, 255, [3] + list(video_resolution), dtype=np.uint8)
            self.video_depth = 3

        self.video_height = video_resolution[0]
        self.video_width = video_resolution[1]

        super().__init__(hero_keys=["pov"], univ_keys=["pov"], space=space)

    def from_hero(self, obs):
        byte_array = super().from_hero(obs)
        pov = np.frombuffer(byte_array, dtype=np.uint8)

        if pov is None or len(pov) == 0:
            pov = np.zeros(
                (self.video_height, self.video_width, self.video_depth), dtype=np.uint8
            ).transpose((2, 0, 1))
        else:
            pov = pov.reshape((self.video_height, self.video_width, self.video_depth))[
                ::-1, :, :
            ].transpose((2, 0, 1))

        return pov

    def __or__(self, other):
        """
        Combines two POV observations into one. If all of the properties match return self
        otherwise raise an exception.
        """
        if (
            isinstance(other, POVObservation)
            and self.include_depth == other.include_depth
            and self.video_resolution == other.video_resolution
        ):
            return POVObservation(
                self.video_resolution, include_depth=self.include_depth
            )
        else:
            raise ValueError("Incompatible observables!")



from typing import Union, Optional, List, Dict, Tuple

from .base import MetaTaskBase
from ...sim.inventory import InventoryItem
from .utils import survive_per_day_reward, survive_n_days_reward, time_since_death_check


class SurvivalMeta(MetaTaskBase):
    """
    Class for survival tasks.
    Args:
        break_speed_multiplier: Controls the speed of breaking blocks. A value larger than 1.0 accelerates the breaking.
                Default: ``1.0``.

        event_level_control: If ``True``, the agent is able to perform high-level controls including place and equip.
                If ``False``, then is keyboard-mouse level control.
                Default: ``True``.

        fast_reset: If ``True``, reset using MC native command `/kill`, instead of waiting for re-generating new worlds.
            Default: ``True``.

            .. warning::
                Side effects:

                1. Changes to the world will not be reset. E.g., if the agent chops lots of trees then calling
                fast reset will not restore those trees.

                2. If you specify agent starting health and food, these specs will only be respected at the first reset
                (i.e., generating a new world) but will not be respected in the following resets (i.e., reset using MC cmds).
                So be careful to use this wrapper if your usages require specific initial health/food.

                3. Statistics/achievements will not be reset. This wrapper will maintain a property ``info_prev_reset``.
                If your tasks use stat/achievements to evaluation, please retrieve this property and compute differences.

        image_size: The size of image observations.

        initial_weather: If not ``None``, specifies the initial weather.
                Can be one of ``"clear"``, ``"normal"``, ``"rain"``, ``"thunder"``.
                Default: ``None``.

        lidar_rays: Defines the directions and maximum distances of the lidar rays if ``use_lidar`` is ``True``.
                If supplied, should be a list of tuple(pitch, yaw, distance).
                Pitch and yaw are in radians and relative to agent looking vector.
                Default: ``None``.

        per_day_reward: The reward value for each day of survival
                Default: ``1``.

        seed: The seed for an instance's internal generator.
                Default: ``None``.

        sim_name: Name of a simulation instance.
                Default: ``"SurvivalMeta"``.

        start_food: If not ``None``, specifies initial food condition of the agent.
                Default: ``None``.

        start_health: If not ``None``, specifies initial health condition of the agent.
                Default: ``None``.

        start_position: If not ``None``, specifies the agent's initial location and orientation in the minecraft world.
                Default: ``None``.

        success_reward: The final reward value if survived target days.
                Default: ``None``.

        target_days: Target days to survive.

        use_lidar: If ``True``, includes lidar in observations.
                Default: ``False``.

        use_voxel: If ``True``, includes voxel in observations.
                Default: ``False``.

        voxel_size: Defines the voxel's range in each axis if ``use_voxel`` is ``True``.
                If supplied, should be a dict with keys ``xmin``, ``xmax``, ``ymin``, ``ymax``, ``zmin``, ``zmax``.
                Each value specifies the voxel size relative to the agent.
                Default: ``None``.

        world_seed: Seed for deterministic world generation
                if ``generate_world_type`` is ``"default"`` or ``"specified_biome"``.
                See `here <https://minecraft.fandom.com/wiki/Seed_(level_generation)>`_ for more details.
                Default: ``None``.
    """

    _prompt_template = "Survive for {target} days."

    # https://minecraft.fandom.com/el/wiki/Day-night_cycle
    mc_ticks_per_day = 24000

    def __init__(
        self,
        *,
        # ------ target days and reward per day
        target_days: int,
        per_day_reward: Union[int, float] = 1,
        success_reward: Optional[Union[int, float]] = None,
        # ------ initial conditions ------
        start_position: Optional[Dict[str, Union[float, int]]] = None,
        start_health: Optional[float] = None,
        start_food: Optional[int] = None,
        initial_weather: Optional[str] = None,
        initial_inventory: Optional[List[InventoryItem]] = None,
        # ------ global conditions ------
        break_speed_multiplier: float = 1.0,
        # ------ sim seed & world seed ------
        seed: Optional[int] = None,
        world_seed: Optional[str] = None,
        # ------ reset mode ------
        fast_reset: bool = True,
        # ------ obs ------
        image_size: Union[int, Tuple[int, int]],
        use_voxel: bool = False,
        voxel_size: Optional[Dict[str, int]] = None,
        use_lidar: bool = False,
        lidar_rays: Optional[List[Tuple[float, float, float]]] = None,
        # ------ event-level action or keyboard-mouse level action ------
        event_level_control: bool = True,
        # ------ misc ------
        sim_name: str = "SurvivalMeta",
    ):
        assert (
            target_days > 1
        ), f"expect target_days > 1 for non-trivial survival task, got {target_days}"
        self._target_days = target_days
        success_reward = success_reward or target_days * per_day_reward

        reward_fns = [
            survive_per_day_reward(self.mc_ticks_per_day, per_day_reward),
            survive_n_days_reward(self.mc_ticks_per_day, target_days, success_reward),
        ]
        success_criteria = [time_since_death_check(self.mc_ticks_per_day * target_days)]

        super().__init__(
            fast_reset=fast_reset,
            success_criteria=success_criteria,
            reward_fns=reward_fns,
            seed=seed,
            sim_name=sim_name,
            image_size=image_size,
            use_voxel=use_voxel,
            voxel_size=voxel_size,
            use_lidar=use_lidar,
            lidar_rays=lidar_rays,
            event_level_control=event_level_control,
            initial_inventory=initial_inventory,
            break_speed_multiplier=break_speed_multiplier,
            world_seed=world_seed,
            start_position=start_position,
            initial_weather=initial_weather,
            start_health=start_health,
            start_food=start_food,
        )

    @property
    def task_prompt(self) -> str:
        return super().get_prompt(target=self._target_days)



from typing import NamedTuple, Optional, Dict, Union, Tuple

from minedojo.sim.mc_meta.mc import ALL_ITEMS


EQUIP_SLOTS = {
    "mainhand": 0,
    "offhand": 40,
    "head": 39,
    "chest": 38,
    "legs": 37,
    "feet": 36,
}
MIN_SLOT_IDX = 0
MAX_SLOT_IDX = 40
SLOT_IDX2NAME = {v: k for k, v in EQUIP_SLOTS.items()}


class InventoryItem(NamedTuple):
    # slot of the item occupies,
    # can be str specifying "main_hand", "off_hand", "head", etc... or int specifying the slot index
    slot: Union[str, int]
    name: str
    variant: Optional[int] = None
    quantity: Optional[int] = None


def parse_inventory_item(item: InventoryItem) -> Tuple[int, Dict[str, Union[str, int]]]:
    slot, name, variant, quantity = item.slot, item.name, item.variant, item.quantity
    if isinstance(slot, str):
        assert slot in EQUIP_SLOTS, f"Unknown slot {slot}"
        slot = EQUIP_SLOTS[slot]
    assert MIN_SLOT_IDX <= slot <= MAX_SLOT_IDX, f"invalid slot index {slot}"
    assert name in ALL_ITEMS, f"Unknown item {name}"
    variant = variant or 0
    quantity = quantity or 1
    return slot, {
        "type": name,
        "metadata": variant,
        "quantity": quantity,
    }


def map_slot_number_to_cmd_slot(slot_number: int) -> str:
    assert MIN_SLOT_IDX <= slot_number <= MAX_SLOT_IDX, "exceed slot index range"
    if slot_number in {0, 40}:
        return f"slot.weapon.{SLOT_IDX2NAME[slot_number]}"
    elif 36 <= slot_number <= 39:
        return f"slot.armor.{SLOT_IDX2NAME[slot_number]}"
    elif 1 <= slot_number <= 8:
        return f"slot.hotbar.{slot_number}"
    else:
        return f"slot.inventory.{slot_number - 9}"



import collections
from typing import Tuple, Optional, Sequence, Dict, Set, List


def decode_item_maybe_with_metadata(s: str) -> Tuple[str, Optional[int]]:
    assert len(s) > 0
    if "#" in s:
        item_type, metadata_str = s.split("#")
        if len(item_type) == 0:
            raise ValueError(f"invalid item_type '{item_type}' in '{s}'")
        metadata = int(metadata_str)
        if metadata not in range(16):
            raise ValueError(f"invalid metadata {metadata} in '{s}'")
        return item_type, int(metadata_str)
    return s, None


def encode_item_with_metadata(item_type: str, metadata: Optional[int]) -> str:
    assert len(item_type) > 0
    if metadata is None:
        return item_type
    else:
        assert isinstance(metadata, int)
        return f"{item_type}#{metadata}"


def error_on_malformed_item_list(
    item_list: Sequence[str], special_items: Sequence[str]
):
    """Check that there are no duplicate item IDs, overlapping item IDs
    (e.g. "planks#2" and "planks"), that special item types don't come with metadata
    constraints (e.g. "air#0" should not be valid).
    """
    map_type_to_metadata: Dict[str, Set[Optional[int]]] = collections.defaultdict(set)
    for s in item_list:
        item_type, metadata = decode_item_maybe_with_metadata(s)
        if item_type in special_items and metadata is not None:
            raise ValueError(
                f"Non-None metadata={metadata} is not allowed for special item type '{item_type}'"
            )

        metadata_set = map_type_to_metadata[item_type]
        if metadata in metadata_set:
            raise ValueError(f"Duplicate item entry for item '{s}'")
        map_type_to_metadata[item_type].add(metadata)

    for item_type, metadata_set in map_type_to_metadata.items():
        if None in metadata_set and len(metadata_set) != 1:
            raise ValueError(
                f"Item list overlaps for item_type={item_type}. This item list includes "
                "both the wildcard metadata option and at least one specific metadata: "
                f"{[n for n in metadata_set if n is not None]}"
            )


def item_list_contains(
    item_list: Sequence[str], item_type: str, metadata: Optional[str]
):
    # log clobber not supported here. (Only used by handlers without log clobber so far).
    if metadata is None:
        return item_type in item_list
    else:
        return encode_item_with_metadata(item_type, metadata) in item_list


def get_unique_matching_item_list_id(
    item_list: Sequence[str],
    item_type: str,
    metadata: int,
    clobber_logs=True,
) -> Optional[str]:
    """
    Assuming that `item_list` doesn't have duplicate elements, returns the string
    identifier in `item_list` that corresponds to `item_type` and `metadata`.

    If there is no matching strings, returns None.

    If there are multiple matching strings, raises a ValueError, because item_lists should
    not have overlapping item identifiers.

    Args:
        item_list: A list of item identifiers. Either just the item type ("wooden_pickaxe")
            or the item type with a metadata requirement ("planks#2").
        item_type: The item type to search for.
        metadata: The metadata to search for.
        clobber_logs: If True, and the only ID in `item_list` starting with "log.*" is the
            ID "log" (no metadata suffix), then treat "log" as a match for "log2". This is
            useful for MineRLTreeChop-v0 where obtaining acacia logs ("log2") should also
            count as obtaining regular logs. Note that no clobbering will occur if any
            log item IDs include metadata, or there's any log2 item IDs in the `item_list`.

            A potentially less tricky solution than this clobbering
            behavior could be adding a special type "log*" type which matches both "log"
            and "log2".
    """
    assert metadata is not None
    if clobber_logs and item_type == "log2":
        log_start = [x for x in item_list if x.startswith("log")]
        if len(log_start) == 1 and log_start[0] == "log":
            item_type = "log"

    result = None
    matches = 0
    if item_type in item_list:
        result = item_type
        matches += 1

    key = encode_item_with_metadata(item_type, metadata)
    if key in item_list:
        result = key
        matches += 1

    if matches > 1:
        raise ValueError(
            f"Multiple item identifiers match with {(item_type, metadata)}"
        )

    return result


def inventory_start_spec_to_item_ids(inv_spec: Sequence[dict]) -> List[str]:
    """Converts the argument of SimpleInventoryAgentStart into a list of equivalent
    item ids suitable for passing into other handlers, like FlatInventoryObservation and
    EquipAction.

    [dict(type=planks, metadata=2, quantity=3),
     dict(type=wooden_pickaxe, quantity=1), ...] => ["planks#2", "wooden_pickaxe", ...]
    """
    result = []
    for d in inv_spec:
        item_type = d["type"]
        metadata = d.get("metadata")
        item_id = encode_item_with_metadata(item_type, metadata)
        result.append(item_id)
    return list(set(result))



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import typing
from typing import Any, Sequence
from collections.abc import Iterable

import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers import util
from minedojo.sim.handlers.translation import TranslationHandler


class Action(TranslationHandler):
    """
    An action handler based on commands
    # Todo: support blacklisting commands. (note this has to work with merging somehow)
    """

    def __init__(self, command: str, space: spaces.MineRLSpace):
        """
        Initializes the space of the handler with a gym.spaces.Dict
        of all of the spaces for each individual command.
        """
        self._command = command
        super().__init__(space)

    @property
    def command(self):
        return self._command

    def to_string(self):
        return self._command

    def to_hero(self, x):
        """
        Returns a command string for the multi command action.
        :param x:
        :return:
        """
        cmd = ""
        verb = self.command

        if isinstance(x, np.ndarray):
            flat = x.flatten().tolist()
            flat = [str(y) for y in flat]
            adjective = " ".join(flat)
        elif isinstance(x, Iterable) and not isinstance(x, str):
            adjective = " ".join([str(y) for y in x])
        else:
            adjective = str(x)
        cmd += "{} {}".format(verb, adjective)

        return cmd

    def __or__(self, other):
        if not self.command == other.command:
            raise ValueError(
                f"Command must be the same between {self.command} and {other.command}"
            )
        return self


class BaseItemListAction(Action):
    def __init__(
        self,
        command: str,
        items: Sequence[str],
        _default="none",
        _other="other",
    ):
        """
        Initializes the space of the handler with a gym.spaces.Dict
        of all of the spaces for each individual command.

        Args:
            command: The 'verb' argument processed by Malmo handlers. This is also
                the default key used by this action in the MineRL env's Dict
                action space.
            items: A list of string item IDs. For convenience, dictionaries in the
                style of those used by `handlers.InventoryAgentStart` can also
                be passed in. In the case of dictionaries, the "type" key is
                unpacked to get the item ID.
        """
        self._items = list(items)
        self._univ_items = ["minecraft:" + item for item in items]
        if _other not in self._items:
            self._items.append(_other)
        if _default not in self._items:
            self._items.append(_default)
        self._default = _default
        self._other = _other
        super().__init__(
            command,
            spaces.Enum(*self._items, default=self._default),
        )

    def __or__(self, other):
        """
        Merges two ItemListCommandActions into one by unioning their items.
        Assert that the commands are the same.
        """

        if not isinstance(other, self.__class__):
            raise TypeError("other must be an instance of ItemListCommandAction")

        if self._command != other._command:
            raise ValueError("Command must be the same for merging")

        new_items = list(set(self._items) | set(other._items))
        return self.__class__(
            command=self._command,
            items=new_items,
            _default=self._default,
            _other=self._other,
        )

    def __eq__(self, other):
        """
        Asserts equality between item list command actions.
        """
        if not isinstance(other, type(self)):
            return False
        if self._command != other._command:
            return False
        if sorted(self._items) != sorted(other._items):
            return False

        return True

    def from_hero(self, x: typing.Dict[str, Any]):
        pass

    def xml_template(self) -> str:
        pass

    @property
    def items(self):
        return self._items

    @property
    def universal_items(self):
        return self._univ_items

    @property
    def default(self):
        return self._default

    def from_universal(self, x):
        raise NotImplementedError


class ItemListAction(BaseItemListAction):
    """
    An action handler based on a list of item IDs without metadata constraints.
    The action space is determined by the length of the list plus one

    Eventually, we should transitions all subclasses of this class to use
    `ItemWithMetadataListAction` instead so that all action item spaces can include
    metadata-constrained item IDs.
    """

    def __init__(self, command: str, items: list, **kwargs):
        """
        Initializes the space of the handler with a gym.spaces.Dict
        of all spaces for each individual command.
        """
        assert all(["#" not in item for item in items])
        super().__init__(command, items, **kwargs)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}" f"(command={self.command}, items={self.items})"
        )


class ItemWithMetadataListAction(BaseItemListAction):
    """
    An action handler based on a list of item IDs with metadata constraints.
    The action space is determined by the length of the list plus one
    """

    def __init__(self, command, items, **kwargs):
        super().__init__(command, items, **kwargs)
        util.error_on_malformed_item_list(items, [self._other, self._default, "air"])

    def _preprocess_item_id(self, item_id: str) -> str:
        """
        Validate item type and metadata (i.e. errors if item_id is malformed), and
        returns an item_id in canonical form. The return value should be the same as the
        argument unless the item has no metadata constraint, in which case "#?" will be
        appended to the argument.
        """
        # Also validates item_id.
        item_type, metadata = util.decode_item_maybe_with_metadata(item_id)

        if not util.item_list_contains(self.items, item_type, metadata):
            raise ValueError(f"{item_id} is not found in {self.items}")
        return util.encode_item_with_metadata(item_type, metadata)

    def to_hero(self, x):
        canonical_item_id = self._preprocess_item_id(x)
        return super().to_hero(canonical_item_id)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from abc import ABC, abstractmethod

import jinja2


class Handler(ABC):
    """Defines the minimal interface for a MineRL handler.

    At their core, handlers should specify unique identifiers
    and a method for producing XML to be given in a mission XML.
    """

    @abstractmethod
    def to_string(self) -> str:
        """The unique identifier for the agent handler.
        This is used for constructing action/observation spaces
        and unioning different env specifications.
        """
        raise NotImplementedError()

    # @abstractmethod #TODO: This should be abstract per convention
    # but this strict handler -> xml enforcement will happen
    # with a pyxb update.
    def xml_template(self) -> str:
        """Generates an XML representation of the handler.

        This XML representation is templated via Jinja2 and
        has access to all of the member variables of the class.

        Note: This is not an abstract method so that
        handlers without corresponding XML's can be combined in
        handler groups with group based XML implementations.
        """
        raise NotImplementedError()

    def xml(self) -> str:
        """Gets the XML representation of Handler by templating
        according to the xml_template class.


        Returns:
            str: the XML representation of the handler.
        """
        var_dict = {}
        for attr_name in dir(self):
            if "xml" not in attr_name:
                var_dict[attr_name] = getattr(self, attr_name)
        try:
            env = jinja2.Environment(undefined=jinja2.StrictUndefined)
            template = env.from_string(self.xml_template())
            return template.render(var_dict)
        except jinja2.UndefinedError as e:
            # print the exception with traceback
            message = e.message + f"\nOccurred in {self}"
            raise jinja2.UndefinedError(message=message)
            pass

    def __or__(self, other):
        """
        Checks to see if self and other have the same to_string
        and if so returns self, otherwise raises an exception.
        """
        assert (
            self.to_string() == other.to_string()
        ), f"Incompatible handlers: {self} and {other}"
        return self

    def __eq__(self, other):
        """
        Checks to see if self and other have the same to_string
        and if so returns self, otherwise raises an exception.
        """
        return self.to_string() == other.to_string()

    def __repr__(self):
        return super().__repr__() + ":" + self.to_string()



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
"""
Server start handlers allow you to set the initial state of the World (e.g. weather, time) 

When used to create a Gym environment, they should be passed to :code:`create_server_initial_conditions`
"""

#  <BiomeGenerator forceReset="true" biome="3"/>
from minedojo.sim.handler import Handler


# <Time>
#     <StartTime>6000</StartTime>
#     <AllowPassageOfTime>false</AllowPassageOfTime>
# </Time>
class TimeInitialCondition(Handler):
    """
    Sets the initial world time as well as whether time can pass.

    Example usage:

    .. code-block:: python

        # Sets time to morning and stops passing of time
        TimeInitialCondition(False, 23000)
    """

    def to_string(self) -> str:
        return "time_initial_condition"

    def xml_template(self) -> str:
        return str(
            """<Time>
                   {% if start_time is not none %}
                   <StartTime>{{start_time | string}}</StartTime>
                   {% endif %}
                   <AllowPassageOfTime>{{allow_passage_of_time | string | lower}}</AllowPassageOfTime>
                </Time>"""
        )

    def __init__(self, allow_passage_of_time: bool, start_time: int = None):
        self.start_time = start_time
        self.allow_passage_of_time = allow_passage_of_time


# <Weather>clear</Weather>
class WeatherInitialCondition(Handler):
    """
    Sets the initial weather condition in the world.

    Example usage:

    .. code-block:: python

        # Sets weather to thunder
        WeatherInitialCondition("thunder")

    """

    def to_string(self) -> str:
        return "weather_initial_condition"

    def xml_template(self) -> str:
        return str("""<Weather>{{weather | string }}</Weather>""")

    def __init__(self, weather: str):
        self.weather = weather


# <AllowSpawning>false</AllowSpawning>
class SpawningInitialCondition(Handler):
    """
    Note: This controls the spawning of MOBs instead of players!
    """

    def to_string(self) -> str:
        return "spawning_initial_condition"

    def xml_template(self) -> str:
        return str(
            """<AllowSpawning>{{allow_spawning | string | lower}}</AllowSpawning>"""
        )

    def __init__(self, allow_spawning: bool):
        self.allow_spawning = allow_spawning
